#include <scheme/GKSNavier.hpp>

#include <language/utils/InterpolateItemValue.hpp>
#include <mesh/ItemValueUtils.hpp>
#include <mesh/ItemValueVariant.hpp>
#include <mesh/MeshTraits.hpp>
#include <mesh/MeshVariant.hpp>
#include <mesh/SubItemValuePerItemVariant.hpp>
#include <scheme/DiscreteFunctionP0.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <utils/Socket.hpp>

#include <variant>
#include <vector>

double
gks_inv_dt(const std::shared_ptr<const DiscreteFunctionVariant>& c_v,
           const std::shared_ptr<const DiscreteFunctionVariant>& U_v)
{
  const auto& c = c_v->get<DiscreteFunctionP0<const double>>();
  const auto& U = U_v->get<DiscreteFunctionP0<const double>>();

  return std::visit(
    [&](auto&& p_mesh) -> double {
      const auto& mesh = *p_mesh;

      using MeshType = decltype(mesh);
      if constexpr (is_polygonal_mesh_v<MeshType>) {
        const auto Vj = MeshDataManager::instance().getMeshData(mesh).Vj();
        const auto Sj = MeshDataManager::instance().getMeshData(mesh).sumOverRLjr();

        CellValue<double> local_inv_dt{mesh.connectivity()};
        parallel_for(
          mesh.numberOfCells(),
          PUGS_LAMBDA(CellId cell_id) { local_inv_dt[cell_id] = (c[cell_id] + abs(U[cell_id])) / (Vj[cell_id]); });

        return max(local_inv_dt);
      } else {
        throw NormalError("unexpected mesh type");
      }
    },
    c.meshVariant()->variant());
}

template <MeshConcept MeshType>
class GKSHandler::GKSNAVIER final : public GKSHandler::IGKSNAVIER
{
 private:
  static constexpr size_t Dimension   = MeshType::Dimension;
  static constexpr size_t SIZEproblem = 2 + Dimension;

  using Rd   = TinyVector<Dimension>;
  using Rdxd = TinyMatrix<Dimension>;

  using MeshDataType = MeshData<MeshType>;

  using DiscreteScalarFunction = DiscreteFunctionP0<const double>;
  using DiscreteVectorFunction = DiscreteFunctionP0<const Rd>;

  const double pi    = std::acos(-1);
  const double sqrt2 = std::sqrt(2);

  TinyMatrix<SIZEproblem>
  _computeInvM_bis(const double& rho,
                   const Rd& rhoU,
                   const double& rhoE,
                   const Rd& U,
                   const double& lambda,
                   const double& delta) const
  {
    TinyMatrix<SIZEproblem> M;
    double U_2 = U[0] * U[0];
    double p   = rho / (2 * lambda);

    M(0, 0) = rho;
    M(1, 1) = rho * U_2 + p;
    M(2, 2) =
      0.25 * rho * U_2 * U_2 + 0.5 * (delta + 3) * U_2 * p + 0.125 * (delta * delta + 4 * delta + 3) * p / lambda;

    M(1, 0) = rhoU[0];
    M(0, 1) = M(1, 0);

    M(2, 0) = rhoE;
    M(0, 2) = M(2, 0);

    M(2, 1) = 0.5 * (rhoU[0] * U_2 + (delta + 3) * U[0] * p);
    M(1, 2) = M(2, 1);

    const TinyMatrix<SIZEproblem> inverseM = inverse(M);

    return inverseM;
  }

  TinyMatrix<SIZEproblem>
  _computeInvMatrixM(const double& rho, const Rd& U, const double& lambda, const double& delta) const
  {
    const double sqrt_rho = std::sqrt(rho);

    const double sqrt_lambda = std::sqrt(lambda);
    const double U_2         = U[0] * U[0];

    TinyMatrix<SIZEproblem> invMatrixM(zero);
    TinyMatrix<SIZEproblem> MatrixP(zero);
    TinyMatrix<SIZEproblem> invMatrixMtilde(zero);

    invMatrixMtilde(0, 0) = 1;
    invMatrixMtilde(1, 1) = 1;
    invMatrixMtilde(2, 2) = (8 * lambda * lambda) / (rho * (delta + 1));

    MatrixP(0, 0) = 1 / sqrt_rho;
    MatrixP(1, 1) = sqrt2 * sqrt_lambda / sqrt_rho;
    MatrixP(2, 2) = 1;
    MatrixP(1, 0) = -sqrt2 * sqrt_lambda * U[0] / sqrt_rho;
    MatrixP(2, 0) = 0.5 * U_2 - 0.25 * (delta + 1) / lambda;
    MatrixP(2, 1) = -U[0];

    invMatrixM = transpose(MatrixP) * invMatrixMtilde * (MatrixP);

    return invMatrixM;
  }

  TinyVector<7>
  _compute_u_moments(const double& rho, const Rd& U, const double& lambda) const
  {
    const double U_2      = U[0] * U[0];
    const double U_4      = U_2 * U_2;
    const double lambda_2 = lambda * lambda;

    TinyVector<7> u_moments;
    u_moments[0] = rho;
    u_moments[1] = rho * U[0];
    u_moments[2] = rho * (U_2 + 0.5 / lambda);
    u_moments[3] = rho * U[0] * (U_2 + 1.5 / lambda);
    u_moments[4] = rho * (U_4 + 3 * U_2 / lambda + 0.75 / lambda_2);
    u_moments[5] = rho * U[0] * (U_4 + 5 * U_2 / lambda + 3.75 / lambda_2);
    u_moments[6] = rho * (U_4 * U_2 + 7.5 * U_4 / lambda + 11.25 * U_2 / lambda_2 + 1.875 / (lambda_2 * lambda));
    return u_moments;
  }

  TinyVector<2>
  _compute_xi2_moments(const double& lambda, const double& delta) const
  {
    TinyVector<2> xi2_moments;
    xi2_moments[0] = delta * 0.5 / lambda;
    xi2_moments[1] = 0.25 * (delta * delta + 2 * delta) / (lambda * lambda);

    return xi2_moments;
  }

  CellValue<const TinyVector<SIZEproblem>>
  _computeGradConservVariables_for_a(const MeshType& mesh,
                                     const DiscreteScalarFunction& rho,
                                     const DiscreteVectorFunction& rhoU,
                                     const DiscreteScalarFunction& rhoE) const
  {
    auto& mesh_data = MeshDataManager::instance().getMeshData(mesh);

    auto Vj = mesh_data.Vj();

    auto node_to_cell_matrix = mesh.connectivity().nodeToCellMatrix();

    CellValue<TinyVector<SIZEproblem>> Gradj{mesh.connectivity()};
    Gradj.fill(zero);

    NodeValue<double> rho_node_mean{mesh.connectivity()};
    NodeValue<Rd> rhoU_node_mean{mesh.connectivity()};
    NodeValue<double> rhoE_node_mean{mesh.connectivity()};
    rho_node_mean.fill(0);
    rhoU_node_mean.fill(zero);
    rhoE_node_mean.fill(0);

    parallel_for(
      mesh.numberOfNodes(), PUGS_LAMBDA(NodeId node_id) {
        const auto& node_cells = node_to_cell_matrix[node_id];

        for (size_t l = 0; l < node_cells.size(); l++) {
          if (node_cells.size() == 1)
            continue;

          rho_node_mean[node_id] += rho[node_cells[l]];
          rhoU_node_mean[node_id] += rhoU[node_cells[l]];
          rhoE_node_mean[node_id] += rhoE[node_cells[l]];
        }

        rho_node_mean[node_id] *= 0.5;
        rhoU_node_mean[node_id] *= 0.5;
        rhoE_node_mean[node_id] *= 0.5;

        Gradj[node_cells[0]][0] = (rho_node_mean[node_id] - rho[node_cells[0]]) / Vj[node_cells[0]];
        Gradj[node_cells[0]][1] = (rhoU_node_mean[node_id][0] - rhoU[node_cells[0]][0]) / Vj[node_cells[0]];
        Gradj[node_cells[0]][2] = (rhoE_node_mean[node_id] - rhoE[node_cells[0]]) / Vj[node_cells[0]];
      });
    return Gradj;
  }

  TinyVector<SIZEproblem>
  _compute_a_and_b(const TinyMatrix<SIZEproblem>& invMatrixM, const TinyVector<SIZEproblem>& Grad) const
  {
    TinyVector<SIZEproblem> a = 2 * invMatrixM * Grad;
    return a;
  }

  TinyVector<SIZEproblem>
  _compute_A(const TinyMatrix<SIZEproblem>& invMatrixM,
             const TinyMatrix<SIZEproblem>& C1_full,
             const TinyVector<SIZEproblem>& a) const
  {
    TinyVector<SIZEproblem> A = (invMatrixM * C1_full) * a;
    return A;
  }

  TinyVector<SIZEproblem>
  _compute_B(const TinyMatrix<SIZEproblem>& invMatrixM_interface,
             const TinyMatrix<SIZEproblem>& C0_semi_pos,
             const TinyMatrix<SIZEproblem>& C0_semi_neg,
             const TinyVector<SIZEproblem>& al,
             const TinyVector<SIZEproblem>& ar) const
  {
    TinyVector<SIZEproblem> B = invMatrixM_interface * (C0_semi_pos * al + C0_semi_neg * ar);
    return B;
  }

  CellValue<TinyVector<SIZEproblem>>
  _computeGradConservVariables_for_b(const MeshType& mesh,
                                     const DiscreteScalarFunction& rho,
                                     const DiscreteVectorFunction& rhoU,
                                     const DiscreteScalarFunction& rhoE,
                                     const NodeValue<double>& rho_node,
                                     const NodeValue<Rd>& rhoU_node,
                                     const NodeValue<double>& rhoE_node) const
  {
    auto& mesh_data = MeshDataManager::instance().getMeshData(mesh);

    auto Vj = mesh_data.Vj();

    auto node_to_cell_matrix = mesh.connectivity().nodeToCellMatrix();

    CellValue<TinyVector<SIZEproblem>> Gradj{mesh.connectivity()};
    Gradj.fill(zero);

    parallel_for(
      mesh.numberOfNodes(), PUGS_LAMBDA(NodeId node_id) {
        const auto& node_cells = node_to_cell_matrix[node_id];

        Gradj[node_cells[0]][0] = (rho_node[node_id] - rho[node_cells[0]]) / Vj[node_cells[0]];
        Gradj[node_cells[0]][1] = (rhoU_node[node_id][0] - rhoU[node_cells[0]][0]) / Vj[node_cells[0]];
        Gradj[node_cells[0]][2] = (rhoE_node[node_id] - rhoE[node_cells[0]]) / Vj[node_cells[0]];
      });

    return Gradj;
  }

  std::tuple<TinyMatrix<SIZEproblem>, TinyMatrix<SIZEproblem>, TinyMatrix<SIZEproblem>>
  _computeMatricesC0C1C2_semi_moments(const double& rho,
                                      const Rd& U,
                                      const double& lambda,
                                      const TinyVector<7>& u_moments,
                                      const TinyVector<2>& xi2_moments,
                                      const int& sign_u) const
  {
    const double U_2      = U[0] * U[0];
    const double U_4      = U_2 * U_2;
    const double lambda_2 = lambda * lambda;

    const double erf_term = (1 + sign_u * std::erf(std::sqrt(lambda) * U[0]));
    const double exp_term = rho * sign_u * std::exp(-lambda * U_2) / std::sqrt(pi * lambda);

    TinyMatrix<SIZEproblem> MatrixC0(zero);
    MatrixC0(0, 0) = 0.5 * u_moments[0] * erf_term;
    MatrixC0(1, 1) = 0.5 * (u_moments[2] * erf_term + U[0] * exp_term);
    MatrixC0(2, 2) = 0.5 * (u_moments[4] * erf_term + U[0] * (U_2 + 2.5 / lambda) * exp_term) +
                     MatrixC0(0, 0) * xi2_moments[1] + 2 * MatrixC0(1, 1) * xi2_moments[0];
    MatrixC0(2, 2) *= 0.25;

    MatrixC0(1, 0) = 0.5 * (u_moments[1] * erf_term + exp_term);
    MatrixC0(0, 1) = MatrixC0(1, 0);
    MatrixC0(2, 0) = MatrixC0(1, 1) + MatrixC0(0, 0) * xi2_moments[0];
    MatrixC0(2, 0) *= 0.5;
    MatrixC0(0, 2) = MatrixC0(2, 0);
    MatrixC0(2, 1) = 0.5 * (u_moments[3] * erf_term + (U_2 + 1 / lambda) * exp_term) + MatrixC0(1, 0) * xi2_moments[0];
    MatrixC0(2, 1) *= 0.5;
    MatrixC0(1, 2) = MatrixC0(2, 1);

    TinyMatrix<SIZEproblem> MatrixC1(zero);
    MatrixC1(0, 0) = MatrixC0(1, 0);
    MatrixC1(1, 1) = 0.5 * (u_moments[3] * erf_term + (U_2 + 1 / lambda) * exp_term);
    MatrixC1(2, 2) = 0.5 * (u_moments[5] * erf_term + (U_4 + 4.5 * U_2 / lambda + 2 / lambda_2) * exp_term) +
                     MatrixC1(0, 0) * xi2_moments[1] + 2 * MatrixC1(1, 1) * xi2_moments[0];
    MatrixC1(2, 2) *= 0.25;

    MatrixC1(1, 0) = MatrixC0(1, 1);
    MatrixC1(0, 1) = MatrixC1(1, 0);
    MatrixC1(2, 0) = MatrixC1(1, 1) + MatrixC1(0, 0) * xi2_moments[0];
    MatrixC1(2, 0) *= 0.5;
    MatrixC1(0, 2) = MatrixC1(2, 0);
    MatrixC1(2, 1) =
      0.5 * (u_moments[4] * erf_term + U[0] * (U_2 + 2.5 / lambda) * exp_term) + MatrixC1(1, 0) * xi2_moments[0];
    MatrixC1(2, 1) *= 0.5;
    MatrixC1(1, 2) = MatrixC1(2, 1);

    TinyMatrix<SIZEproblem> MatrixC2(zero);
    MatrixC2(0, 0) = MatrixC1(1, 0);
    MatrixC2(1, 1) = 0.5 * (u_moments[4] * erf_term + U[0] * (U_2 + 2.5 / lambda) * exp_term);
    MatrixC2(2, 2) = 0.5 * (u_moments[6] * erf_term + U[0] * (U_4 + 7 * U_2 / lambda + 8.25 / lambda_2) * exp_term) +
                     MatrixC2(0, 0) * xi2_moments[1] + 2 * MatrixC2(1, 1) * xi2_moments[0];
    MatrixC2(2, 2) *= 0.25;

    MatrixC2(1, 0) = MatrixC1(1, 1);
    MatrixC2(0, 1) = MatrixC2(1, 0);
    MatrixC2(2, 0) = MatrixC2(1, 1) + MatrixC2(0, 0) * xi2_moments[0];
    MatrixC2(2, 0) *= 0.5;
    MatrixC2(0, 2) = MatrixC2(2, 0);
    MatrixC2(2, 1) = 0.5 * (u_moments[5] * erf_term + (U_4 + 4.5 * U_2 / lambda + 2 / lambda_2) * exp_term) +
                     MatrixC2(1, 0) * xi2_moments[0];
    MatrixC2(2, 1) *= 0.5;
    MatrixC2(1, 2) = MatrixC2(2, 1);

    return {MatrixC0, MatrixC1, MatrixC2};
  }

  TinyMatrix<SIZEproblem>
  _computeMatricesC1_full_moments(const TinyVector<7>& u_moments, const TinyVector<2>& xi2_moments) const
  {
    TinyMatrix<SIZEproblem> MatrixC1(zero);

    MatrixC1(0, 0) = u_moments[1];
    MatrixC1(1, 1) = u_moments[3];
    MatrixC1(2, 2) = 0.25 * (u_moments[5] + 2 * xi2_moments[0] * u_moments[3] + xi2_moments[1] * u_moments[1]);

    MatrixC1(1, 0) = u_moments[2];
    MatrixC1(0, 1) = MatrixC1(1, 0);
    MatrixC1(2, 0) = 0.5 * (u_moments[3] + u_moments[1] * xi2_moments[0]);
    MatrixC1(0, 2) = MatrixC1(2, 0);
    MatrixC1(2, 1) = (u_moments[4] + u_moments[2] * xi2_moments[0]);
    MatrixC1(1, 2) = MatrixC1(2, 1);

    return {MatrixC1};
  }

 public:
  std::tuple<const std::shared_ptr<const ItemValueVariant>,
             const std::shared_ptr<const ItemValueVariant>,
             const std::shared_ptr<const ItemValueVariant>>
  compute_fluxes(const std::shared_ptr<const DiscreteFunctionVariant>& rho_v,
                 const std::shared_ptr<const DiscreteFunctionVariant>& rhoU_v,
                 const std::shared_ptr<const DiscreteFunctionVariant>& rhoE_v,
                 const double& viscosity,
                 const double& delta,
                 const double& dt) const
  {
    auto mesh_v          = getCommonMesh({rho_v, rhoU_v, rhoE_v});
    const MeshType& mesh = *mesh_v->get<MeshType>();

    DiscreteScalarFunction rho_n  = rho_v->get<DiscreteScalarFunction>();
    DiscreteVectorFunction rhoU_n = rhoU_v->get<DiscreteVectorFunction>();
    DiscreteScalarFunction rhoE_n = rhoE_v->get<DiscreteScalarFunction>();

    DiscreteScalarFunction rho  = copy(rho_n);
    DiscreteVectorFunction rhoU = copy(rhoU_n);
    DiscreteScalarFunction rhoE = copy(rhoE_n);

    CellValue<double> lambda{mesh.connectivity()};
    lambda.fill(1);
    CellValue<Rd> U{mesh.connectivity()};
    U.fill(zero);

    auto& mesh_data = MeshDataManager::instance().getMeshData(mesh);

    auto Vj = mesh_data.Vj();

    auto cell_to_node_matrix = mesh.connectivity().cellToNodeMatrix();
    auto node_to_cell_matrix = mesh.connectivity().nodeToCellMatrix();

    const NodeValuePerCell<const Rd> njr = mesh_data.njr();

    CellValue<double> rho_fluxes(mesh.connectivity());
    CellValue<Rd> rhoU_fluxes(mesh.connectivity());
    CellValue<double> rhoE_fluxes(mesh.connectivity());
    rho_fluxes.fill(0);
    rhoU_fluxes.fill(zero);
    rhoE_fluxes.fill(0);

    NodeValue<double> rho_node(mesh.connectivity());
    NodeValue<Rd> rhoU_node(mesh.connectivity());
    NodeValue<double> rhoE_node(mesh.connectivity());
    NodeValue<Rd> U_node(mesh.connectivity());
    NodeValue<double> lambda_node(mesh.connectivity());
    rho_node.fill(1);
    rhoU_node.fill(zero);
    rhoE_node.fill(1);
    U_node.fill(zero);
    lambda_node.fill(1);

    CellValue<double> err_function_term(mesh.connectivity());
    CellValue<double> exp_term(mesh.connectivity());

    parallel_for(
      mesh.numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
        U[cell_id][0]   = rhoU_n[cell_id][0] / rho_n[cell_id];
        double rhoU_2   = rhoU_n[cell_id][0] * U[cell_id][0];
        lambda[cell_id] = 0.5 * (1. + delta) * rho_n[cell_id] / (2 * rhoE_n[cell_id] - rhoU_2);
      });

    parallel_for(
      mesh.numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
        err_function_term[cell_id] = std::erf(std::sqrt(lambda[cell_id]) * U[cell_id][0]);
        exp_term[cell_id] =
          std::exp(-lambda[cell_id] * U[cell_id][0] * U[cell_id][0]) / std::sqrt(pi * lambda[cell_id]);
      });

    parallel_for(
      mesh.numberOfNodes(), PUGS_LAMBDA(NodeId node_id) {
        const auto& node_cells = node_to_cell_matrix[node_id];

        double rho_cell_left   = 1;
        double rho_cell_right  = 1;
        double rhoU_cell_left  = 1;
        double rhoU_cell_right = 1;
        double rhoE_cell_left  = 1;
        double rhoE_cell_right = 1;

        for (size_t l = 0; l < node_cells.size(); l++) {
          if (node_cells.size() == 1)
            continue;

          if (l == 0) {
            rho_cell_left = rho_n[node_cells[l]] * (1 + err_function_term[node_cells[l]]);

            rhoU_cell_left = rhoU_n[node_cells[l]][0] * (1. + err_function_term[node_cells[l]]) +
                             rho_n[node_cells[l]] * exp_term[node_cells[l]];

            rhoE_cell_left = rhoE_n[node_cells[l]] * (1. + err_function_term[node_cells[l]]) +
                             0.5 * rhoU_n[node_cells[l]][0] * exp_term[node_cells[l]];
          } else {
            rho_cell_right = rho_n[node_cells[l]] * (1 - err_function_term[node_cells[l]]);

            rhoU_cell_right = rhoU_n[node_cells[l]][0] * (1. - err_function_term[node_cells[l]]) -
                              rho_n[node_cells[l]] * exp_term[node_cells[l]];

            rhoE_cell_right = rhoE_n[node_cells[l]] * (1. - err_function_term[node_cells[l]]) -
                              0.5 * rhoU_n[node_cells[l]][0] * exp_term[node_cells[l]];
          }
        }
        rho_node[node_id]     = 0.5 * (rho_cell_left + rho_cell_right);
        rhoU_node[node_id][0] = 0.5 * (rhoU_cell_left + rhoU_cell_right);
        rhoE_node[node_id]    = 0.5 * (rhoE_cell_left + rhoE_cell_right);
      });

    parallel_for(
      mesh.numberOfNodes(), PUGS_LAMBDA(NodeId node_id) {
        U_node[node_id][0]   = rhoU_node[node_id][0] / rho_node[node_id];
        double rhoU2_node    = rhoU_node[node_id][0] * rhoU_node[node_id][0] / rho_node[node_id];
        lambda_node[node_id] = (1 + delta) * rho_node[node_id] / (2 * (2 * rhoE_node[node_id] - rhoU2_node));
      });

    NodeValue<double> tau{mesh.connectivity()};
    tau.fill(1);
    NodeValue<double> eta{mesh.connectivity()};
    eta.fill(1);
    NodeValue<double> zeta{mesh.connectivity()};
    zeta.fill(1);
    for (NodeId node_id = 0; node_id < mesh.numberOfNodes(); ++node_id) {
      tau[node_id] = viscosity * rho_node[node_id] / (2 * lambda_node[node_id]);
      if (viscosity == 0) {
        eta[node_id]  = 0;
        zeta[node_id] = 0;
      } else {
        zeta[node_id] = std::exp(-dt / tau[node_id]);
        eta[node_id]  = (tau[node_id] / dt) * (1 - zeta[node_id]);
      }
    }

    auto Gradj_for_a = this->_computeGradConservVariables_for_a(mesh, rho, rhoU, rhoE);
    auto Gradj_for_b = this->_computeGradConservVariables_for_b(mesh, rho, rhoU, rhoE, rho_node, rhoU_node, rhoE_node);

    NodeValue<TinyVector<SIZEproblem>> Fluxes_term{mesh.connectivity()};
    Fluxes_term.fill(zero);

    parallel_for(
      mesh.numberOfNodes(), PUGS_LAMBDA(NodeId node_id) {
        const auto& node_cells = node_to_cell_matrix[node_id];

        const double rho_left_node = rho[node_cells[0]] + 0.5 * Vj[node_cells[0]] * Gradj_for_a[node_cells[0]][0];
        Rd rhoU_left_node;
        rhoU_left_node[0]           = rhoU[node_cells[0]][0] + 0.5 * Vj[node_cells[0]] * Gradj_for_a[node_cells[0]][1];
        const double rhoE_left_node = rhoE[node_cells[0]] + 0.5 * Vj[node_cells[0]] * Gradj_for_a[node_cells[0]][2];

        Rd U_left_node;
        U_left_node[0] = rhoU_left_node[0] / rho_left_node;
        const double lambda_left_node =
          (1 + delta) * rho_left_node / (2 * (2 * rhoE_left_node - rhoU_left_node[0] * U_left_node[0]));

        const double rho_right_node = rho[node_cells[1]] - 0.5 * Vj[node_cells[1]] * Gradj_for_a[node_cells[1]][0];
        Rd rhoU_right_node;
        rhoU_right_node[0]           = rhoU[node_cells[1]][0] - 0.5 * Vj[node_cells[1]] * Gradj_for_a[node_cells[1]][1];
        const double rhoE_right_node = rhoE[node_cells[1]] - 0.5 * Vj[node_cells[1]] * Gradj_for_a[node_cells[1]][2];

        Rd U_right_node;
        U_right_node[0] = rhoU_right_node[0] / rho_right_node;
        const double lambda_right_node =
          (1 + delta) * rho_right_node / (2 * (2 * rhoE_right_node - rhoU_right_node[0] * U_right_node[0]));

        TinyVector<7> u_moments_left   = this->_compute_u_moments(rho_left_node, U_left_node, lambda_left_node);
        TinyVector<2> xi2_moments_left = this->_compute_xi2_moments(lambda_left_node, delta);

        TinyVector<7> u_moments_right   = this->_compute_u_moments(rho_right_node, U_right_node, lambda_right_node);
        TinyVector<2> xi2_moments_right = this->_compute_xi2_moments(lambda_right_node, delta);

        TinyMatrix<SIZEproblem> InvM_left =
          this->_computeInvMatrixM(rho_left_node, U_left_node, lambda_left_node, delta);
        TinyMatrix<SIZEproblem> InvM_right =
          this->_computeInvMatrixM(rho_right_node, U_right_node, lambda_right_node, delta);
        TinyMatrix<SIZEproblem> InvMInterface =
          this->_computeInvMatrixM(rho_node[node_id], U_node[node_id], lambda_node[node_id], delta);

        // //For the calculation of matrices M using the inverse() routine of pugs for TinyMatrix
        // TinyMatrix<SIZEproblem> InvM_left =
        //   this->_computeInvM_bis(rho_left_node, rhoU_left_node, rhoE_left_node, U_left_node, lambda_left_node,
        //   delta);
        // TinyMatrix<SIZEproblem> InvM_right = this->_computeInvM_bis(rho_right_node, rhoU_right_node, rhoE_right_node,
        //                                                             U_right_node, lambda_right_node, delta);
        // TinyMatrix<SIZEproblem> InvMInterface =
        //   this->_computeInvM_bis(rho_node[node_id], rhoU_node[node_id], rhoE_node[node_id], U_node[node_id],
        //                          lambda_node[node_id], delta);

        TinyVector<SIZEproblem> al;
        TinyVector<SIZEproblem> ar;
        TinyVector<SIZEproblem> bl;
        TinyVector<SIZEproblem> br;

        al = this->_compute_a_and_b(InvM_left, Gradj_for_a[node_cells[0]]);
        bl = this->_compute_a_and_b(InvMInterface, Gradj_for_b[node_cells[0]]);
        ar = this->_compute_a_and_b(InvM_right, Gradj_for_a[node_cells[0]]);
        br = this->_compute_a_and_b(InvMInterface, Gradj_for_b[node_cells[0]]);

        auto [C0_semi_pos_left, C1_semi_pos_left, C2_semi_pos_left] =
          this->_computeMatricesC0C1C2_semi_moments(rho_left_node, U_left_node, lambda_left_node, u_moments_left,
                                                    xi2_moments_left, 1);

        auto [C0_semi_neg_right, C1_semi_neg_right, C2_semi_neg_right] =
          this->_computeMatricesC0C1C2_semi_moments(rho_right_node, U_right_node, lambda_right_node, u_moments_right,
                                                    xi2_moments_left, -1);

        TinyVector<SIZEproblem> Al = this->_compute_A(InvM_left, C1_semi_pos_left, -al);
        TinyVector<SIZEproblem> Ar = this->_compute_A(InvM_right, C1_semi_neg_right, -ar);

        TinyVector<SIZEproblem> B = this->_compute_B(InvMInterface, C0_semi_pos_left, C0_semi_neg_right, al, ar);

        TinyVector<SIZEproblem> Fluxes_from_a = (tau[node_id] * zeta[node_id] - tau[node_id] * eta[node_id]) *
                                                (C2_semi_pos_left * al + C2_semi_neg_right * ar);
        TinyVector<SIZEproblem> Fluxes_for_Navier =
          -tau[node_id] * eta[node_id] *
          ((C1_semi_pos_left * Al + C1_semi_neg_right * Ar) + (C2_semi_pos_left * al + C2_semi_neg_right * ar));

        TinyVector<7> u_moments_interface =
          this->_compute_u_moments(rho_node[node_id], U_node[node_id], lambda_node[node_id]);
        TinyVector<2> xi2_moments_interface = this->_compute_xi2_moments(lambda_node[node_id], delta);

        auto [C0_semi_pos_interface, C1_semi_pos_interface, C2_semi_pos_interface] =
          this->_computeMatricesC0C1C2_semi_moments(rho_node[node_id], U_node[node_id], lambda_node[node_id],
                                                    u_moments_interface, xi2_moments_interface, 1);
        auto [C0_semi_neg_interface, C1_semi_neg_interface, C2_semi_neg_interface] =
          this->_computeMatricesC0C1C2_semi_moments(rho_node[node_id], U_node[node_id], lambda_node[node_id],
                                                    u_moments_interface, xi2_moments_interface, -1);

        TinyMatrix<SIZEproblem> Matrix_for_B =
          this->_computeMatricesC1_full_moments(u_moments_interface, xi2_moments_interface);

        TinyVector<SIZEproblem> Fluxes_from_b = (2 * tau[node_id] * eta[node_id] - tau[node_id] * (1 + zeta[node_id])) *
                                                (C2_semi_pos_interface * bl + C2_semi_neg_interface * br);
        TinyVector<SIZEproblem> Fluxes_from_B =
          (dt / 2 - tau[node_id] + tau[node_id] * eta[node_id]) * Matrix_for_B * B;

        TinyVector<SIZEproblem> G;
        G[0] = u_moments_interface[1];
        G[1] = u_moments_interface[2];
        G[2] = 0.5 * (u_moments_interface[3] + u_moments_interface[1] * xi2_moments_interface[0]);

        TinyVector<SIZEproblem> F_fn;
        F_fn[0] = C1_semi_pos_left(0, 0) + C1_semi_neg_right(0, 0);
        F_fn[1] = C1_semi_pos_left(1, 0) + C1_semi_neg_right(1, 0);
        F_fn[2] = C1_semi_pos_left(2, 0) + C1_semi_neg_right(2, 0);

        Fluxes_term[node_id] = (1 - eta[node_id]) * G + eta[node_id] * F_fn + Fluxes_for_Navier + Fluxes_from_a +
                               Fluxes_from_b + Fluxes_from_B;
      });

    NodeValuePerCell<TinyVector<SIZEproblem>> Fluxes_term_apply{mesh.connectivity()};
    Fluxes_term_apply.fill(zero);

    parallel_for(
      mesh.numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
        const auto& cell_nodes = cell_to_node_matrix[cell_id];

        for (size_t r = 0; r < cell_nodes.size(); r++) {
          Fluxes_term_apply(cell_id, r) = njr(cell_id, r)[0] * Fluxes_term[cell_nodes[r]];
        }
      });

    for (CellId cell_id = 2; cell_id < mesh.numberOfCells() - 2; ++cell_id) {
      const size_t& nb_nodes = Fluxes_term_apply.numberOfSubValues(cell_id);
      for (size_t r = 0; r < nb_nodes; r++) {
        rho_fluxes[cell_id] += Fluxes_term_apply(cell_id, r)[0];
        rhoU_fluxes[cell_id][0] += Fluxes_term_apply(cell_id, r)[1];
        rhoE_fluxes[cell_id] += Fluxes_term_apply(cell_id, r)[2];
      }
    }

    return std::make_tuple(std::make_shared<const ItemValueVariant>(rho_fluxes),
                           std::make_shared<const ItemValueVariant>(rhoU_fluxes),
                           std::make_shared<const ItemValueVariant>(rhoE_fluxes));
  }

  std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>>
  apply_fluxes(const MeshType& mesh,
               const DiscreteScalarFunction& rho,
               const DiscreteVectorFunction& rhoU,
               const DiscreteScalarFunction& rhoE,
               const CellValue<const double>& rho_fluxes,
               const CellValue<const Rd>& rhoU_fluxes,
               const CellValue<const double>& rhoE_fluxes,
               const double& dt) const
  {
    CellValue<double> new_rho  = copy(rho.cellValues());
    CellValue<Rd> new_rhoU     = copy(rhoU.cellValues());
    CellValue<double> new_rhoE = copy(rhoE.cellValues());

    NodeValue<Rd> xr = copy(mesh.xr());

    std::shared_ptr<const MeshType> mesh_v = std::make_shared<MeshType>(mesh.shared_connectivity(), xr);

    auto& mesh_data = MeshDataManager::instance().getMeshData(mesh);
    auto Vj         = mesh_data.Vj();

    for (CellId cell_id = 2; cell_id < mesh.numberOfCells() - 2; ++cell_id) {
      new_rho[cell_id] -= dt / Vj[cell_id] * rho_fluxes[cell_id];
      new_rhoU[cell_id] -= dt / Vj[cell_id] * rhoU_fluxes[cell_id];
      new_rhoE[cell_id] -= dt / Vj[cell_id] * rhoE_fluxes[cell_id];
    }

    return {std::make_shared<DiscreteFunctionVariant>(DiscreteScalarFunction(mesh_v, new_rho)),
            std::make_shared<DiscreteFunctionVariant>(DiscreteVectorFunction(mesh_v, new_rhoU)),
            std::make_shared<DiscreteFunctionVariant>(DiscreteScalarFunction(mesh_v, new_rhoE))};
  }

  std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>>
  apply_fluxes(const std::shared_ptr<const DiscreteFunctionVariant>& rho_v,
               const std::shared_ptr<const DiscreteFunctionVariant>& rhoU_v,
               const std::shared_ptr<const DiscreteFunctionVariant>& rhoE_v,
               const std::shared_ptr<const ItemValueVariant>& rho_fluxes,
               const std::shared_ptr<const ItemValueVariant>& rhoU_fluxes,
               const std::shared_ptr<const ItemValueVariant>& rhoE_fluxes,
               const double& dt) const
  {
    std::shared_ptr mesh_v = getCommonMesh({rho_v, rhoU_v, rhoE_v});
    if (not mesh_v) {
      throw NormalError("discrete functions are not defined on the same mesh");
    }

    if (not checkDiscretizationType({rho_v, rhoU_v, rhoE_v}, DiscreteFunctionType::P0)) {
      throw NormalError("GKS solver expects P0 functions");
    }

    return this->apply_fluxes(*mesh_v->get<MeshType>(),                      //
                              rho_v->get<DiscreteScalarFunction>(),          //
                              rhoU_v->get<DiscreteVectorFunction>(),         //
                              rhoE_v->get<DiscreteScalarFunction>(),         //
                              rho_fluxes->get<CellValue<const double>>(),    //
                              rhoU_fluxes->get<CellValue<const Rd>>(),       //
                              rhoE_fluxes->get<CellValue<const double>>(),   //
                              dt);
  }

  std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,   // rho
             std::shared_ptr<const DiscreteFunctionVariant>,   // rho U
             std::shared_ptr<const DiscreteFunctionVariant>>   // rho E
  gksNavier(const std::shared_ptr<const DiscreteFunctionVariant>& rho_v,
            const std::shared_ptr<const DiscreteFunctionVariant>& rhoU_v,
            const std::shared_ptr<const DiscreteFunctionVariant>& rhoE_v,
            const double& viscosity,
            const double& delta,
            const double& dt) const
  {
    std::shared_ptr mesh_v = getCommonMesh({rho_v, rhoU_v, rhoE_v});

    auto [rho_fluxes, rhoU_fluxes, rhoE_fluxes] = compute_fluxes(rho_v, rhoU_v, rhoE_v, viscosity, delta, dt);
    return apply_fluxes(rho_v, rhoU_v, rhoE_v, rho_fluxes, rhoU_fluxes, rhoE_fluxes, dt);
  }

  GKSNAVIER()            = default;
  GKSNAVIER(GKSNAVIER&&) = default;
  ~GKSNAVIER()           = default;
};

GKSHandler::GKSHandler(const std::shared_ptr<const MeshVariant>& mesh_v)
{
  if (not mesh_v) {
    throw NormalError("discrete functions are not defined on the same mesh");
  }

  std::visit(
    [&](auto&& mesh) {
      using MeshType = mesh_type_t<decltype(mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType>) {
        if constexpr (MeshType::Dimension == 1) {
          m_gks_navier = std::make_unique<GKSNAVIER<MeshType>>();
        } else {
          throw NormalError("unexpected mesh type");
        }
      } else {
        throw NormalError("unexpected mesh type");
      }
    },
    mesh_v->variant());
}
